#version 330
#pragma optionNV (unroll all)

#include "poisson.glsl"
#include "colorSpaces.glsl"
#line 7
#define BLOCKER_SEARCH_NUM_SAMPLES 16
#define PCF_NUM_SAMPLES 32
#define Z_BIAS 0.00215 // against shadow acne. Higher -> less acne & more peter panning!

in vec2 TexCoord0;
in vec2 XYPos0; // = TexCoord0 * 2 - 1

out vec4 FragColor;

uniform sampler2D DiffuseTexture; // diffuse texture from GBuffer
uniform sampler2D NormalTexture; // normal texture from GBuffer
uniform sampler2D MaterialTexture; // material texture from GBuffer
uniform sampler2D DepthTexture; // depth texture from GBuffer
uniform sampler2D OcclusionTexture; // AO texture
uniform sampler2D LightAccumulationTexture; // other light contributions (rgb == diffuse, alpha = specular)
uniform sampler2DShadow ShadowTexture; // shadow map with shadow sampler
uniform sampler2D ShadowTextureNormal; // shadow map with 'normal' depth reads

uniform float LightAmbientPower;
uniform float LightPower;
uniform vec3  LightColor;
uniform vec3  LightDirection_world; // assumed normalized
uniform vec3  EyePosition_world;

uniform mat4 InvViewProjMatrix;
uniform mat4 LightViewProjTexMatrix; // light view projection matrix + mapping from clip space to texture space

uniform bool DisplayAmbientOcclusion;
uniform bool IncludeAmbientOcclusion;
uniform uint ShadowType; // 0 .. None, 1 .. PCSS, 2 .. Simple PCF
uniform float LightSize; // ~ between [0.02, 0.1]
uniform float ShadowBrightness; // higher -> brighter
uniform bool DayMode;

#if BLOCKER_SEARCH_NUM_SAMPLES == 64
	#define POISSON_DISK_BLOCKER poissonDisk64
#elif BLOCKER_SEARCH_NUM_SAMPLES == 32
	#define POISSON_DISK_BLOCKER poissonDisk32
#elif BLOCKER_SEARCH_NUM_SAMPLES == 25
	#define POISSON_DISK_BLOCKER poissonDisk25
#else
	#define POISSON_DISK_BLOCKER poissonDisk16
#endif

#if PCF_NUM_SAMPLES == 64
	#define POISSON_DISK_PCF poissonDisk64
#elif PCF_NUM_SAMPLES == 32
	#define POISSON_DISK_PCF poissonDisk32
#elif PCF_NUM_SAMPLES == 25
	#define POISSON_DISK_PCF poissonDisk25
#else
	#define POISSON_DISK_PCF poissonDisk16
#endif


// Idea from: http://www.derschmale.com/2014/07/24/faster-variance-soft-shadow-mapping-for-varying-penumbra-sizes/
float EstimatePenumbraSize(in float lightSize, float zCoord, in float averageOccluderDepth, in float penumbraScale)
{
	float penumbraSize = lightSize * (zCoord - averageOccluderDepth) * penumbraScale;

	// clamp to the maximum softness, which matches the search area
	return min(penumbraSize, lightSize);
}

void FindBlocker(in vec2 uv, in float zReceiver, in vec2 searchWidth, out float avgBlockerDepth, out float numBlockers)
{
	float blockerSum = 0;
	numBlockers = 0;
	
	// sample the center position as well:
	float shadowMapDepth = texture(ShadowTextureNormal, uv).r;
	if(shadowMapDepth < zReceiver)
	{
		blockerSum += shadowMapDepth;
		++numBlockers;
	}
	
	// poisson sampling for blockers:
	for(int i=0;i<BLOCKER_SEARCH_NUM_SAMPLES;++i)
	{
		float shadowMapDepth = texture(ShadowTextureNormal, uv + POISSON_DISK_BLOCKER[i] * searchWidth).r;
		if(shadowMapDepth < zReceiver)
		{
			blockerSum += shadowMapDepth;
			++numBlockers;
		}
	}
	
	// divide by zero can happen, but is handled in the PCSS function using the numBlockers variable
	avgBlockerDepth = blockerSum / numBlockers;
}

float sampleShadow(in vec3 texC, in vec2 filterRadiusUV)
{
	//return texture(ShadowTexture, vec3(texC.xy, texC.z)); // used for no filtering

	float sum = 0.0;
	for(int i = 0; i < PCF_NUM_SAMPLES; i++)
	{
		vec2 offset = POISSON_DISK_PCF[i] * filterRadiusUV;
		sum += texture(ShadowTexture, vec3(texC.xy + offset, texC.z));
	}
	return sum / PCF_NUM_SAMPLES;
}

float PCSS (in vec3 coords)
{
	float avgBlockerDepth = 0.0f;
	float numBlockers = 0.0f;
	
	// search width estimation is tricky for directional light, so we just use a constant value
	vec2 searchWidth = vec2(1) * LightSize / 2;

	FindBlocker(coords.xy, coords.z, searchWidth, avgBlockerDepth, numBlockers);
	if (numBlockers < 1) return 1.0f;
		
	float penumbraSize = EstimatePenumbraSize(LightSize, coords.z, avgBlockerDepth, 3.8);
	vec2 filterRadiusUV = vec2(penumbraSize / 2.0f);
	
	float shadow = sampleShadow(coords, filterRadiusUV);

	return clamp(ShadowBrightness + shadow, 0.0, 1.0);
}

float SimpleShadowing(in vec3 coords)
{
	vec2 uvCoords = coords.xy;
	float z = coords.z;
  
	ivec2 texSize = textureSize(ShadowTexture, 0);
	float xOffset = 1.0 / texSize.x;
	float yOffset = 1.0 / texSize.y;

	float Factor = 0.0;


	for (int y = -1 ; y <= 1 ; y++) {
		for (int x = -1 ; x <= 1 ; x++) {
			vec2 offsets = vec2(x * xOffset, y * yOffset);
			vec3 UVC = vec3(uvCoords + offsets, z);
			Factor += texture(ShadowTexture, UVC);
		}
	}
	
	return clamp(ShadowBrightness + Factor / 9.0, 0.0, 1.0);
}

vec3 positionFromDepth(in vec2 texCoord, in vec2 xyCoord)
{
	float z = texture(DepthTexture, texCoord).r * 2 - 1; // range between [-1, 1]
	vec4 projectedPos = vec4(xyCoord, z, 1.0f);

	vec4 pos_world_hom = InvViewProjMatrix * projectedPos;
	return pos_world_hom.xyz / pos_world_hom.w;
}

void main()
{
	vec3 position_world = positionFromDepth(TexCoord0, XYPos0);
	vec3 normal_world = texture(NormalTexture, TexCoord0).xyz;
	vec4 diffSpec = texture(DiffuseTexture, TexCoord0);
	vec4 lightAcc = texture(LightAccumulationTexture, TexCoord0);
	float specularPower = clamp(texture(MaterialTexture, TexCoord0).r * 65536.0f, 1.0f, 100.0f);

	vec4 pos_cs = LightViewProjTexMatrix * vec4(position_world, 1.0f); // clip space, shifted by *0.5 + 0.5 (i.e. from [-1,1] to [0,1])
	vec3 pos_ndc = pos_cs.xyz / pos_cs.w;

	pos_ndc.z -= Z_BIAS; // bias

	float shadowing = 1; // 1 = no shadowing
	if (ShadowType == 1u)
	{
		shadowing = PCSS(pos_ndc);
	}
	else if (ShadowType == 2u)
	{
		shadowing = SimpleShadowing(pos_ndc); // for non-soft PCF-filtered shadows (fast)
	}
	// else -> no shadows

	float lightOcclusion = 1;
	if (IncludeAmbientOcclusion)
	{
		vec4 occlusion = texture(OcclusionTexture, TexCoord0);
		lightOcclusion = 1 - clamp(dot(vec4(-LightDirection_world, 1), occlusion), 0.0, 1.0); // 1 if no occlusion
	}

	if (DisplayAmbientOcclusion)
	{
		FragColor = vec4(vec3(lightOcclusion), 1.0f);
	}
	else
	{
		vec3 diffuseAlbedo = diffSpec.xyz;
		vec3 specularAlbedo = vec3(0.3f);
		float specularIntensity = diffSpec.w;

		vec3 L = -LightDirection_world;
		vec3 V = normalize( EyePosition_world - position_world );
		vec3 H = normalize( V + L );

		float NL = clamp(dot(normal_world, L), 0.0f, 1.0f);
		float NH = clamp(dot(normal_world, H), 0.0f, 1.0f);

		vec3 diffuse  = diffuseAlbedo * NL;
		vec3 specular = specularIntensity * specularAlbedo * pow( NH, specularPower );
		vec3 ambient  = diffuseAlbedo * LightAmbientPower; // * max(lightOcclusion, 0.01); // looks better without occluding the ambient part

		vec3 color = ambient + (diffuse + specular) * LightColor * LightPower * lightOcclusion * shadowing;

		vec3 finalColor = color;
		if (!DayMode)
		{
			// scotopic luminance:
			// see http://content.gpwiki.org/index.php/D3DBook:High-Dynamic_Range_Rendering
			// Equations 19, 20
			vec3 XYZ = rgb2xyz_V1(color);

			float Vl = 0;
			if (XYZ.x != 0)
			{
				// avoid NaNs
				Vl = XYZ.y * (1.33 * (1 + (XYZ.y + XYZ.z) / XYZ.x) - 1.68);
			}

			vec3 nightColor = vec3(1.05, 0.97, 1.27) * Vl;
			finalColor = mix(nightColor, color, 0.3f); // linear mix between the two colors
		}

		vec3 additionalDiffuse = lightAcc.rgb;
		float lum = dot(vec3(0.2126f, 0.7152f, 0.0722f), additionalDiffuse);
		vec3 additionalSpecular = additionalDiffuse * lightAcc.a / (lum + 0.001);

		vec3 additionalColor = diffuseAlbedo * additionalDiffuse + specularAlbedo * additionalSpecular;

		//finalColor = mix(finalColor, additionalColor, 0.5f);
		finalColor += additionalColor;

		FragColor = vec4(finalColor, 1.0f);
	}
}
