#version 330

out vec4 FragColor;

uniform sampler2D DiffuseTexture; // diffuse texture from GBuffer
uniform sampler2D NormalTexture; // normal texture from GBuffer
uniform sampler2D MaterialTexture; // material texture from GBuffer
uniform sampler2D DepthTexture; // depth texture from GBuffer
uniform sampler2D OcclusionTexture; // AO texture

uniform float LightPower;
uniform vec3  LightColor;
uniform vec3  LightAttenuation; // x .. constant attenuation, y .. linear, z .. quadratic
uniform float LightRange;
uniform vec3  LightPosition_world;
uniform vec3  EyePosition_world;

uniform mat4 InvViewProjMatrix;
uniform bool IncludeAmbientOcclusion;

vec3 positionFromDepth(in vec2 texCoord)
{
	float z = texture(DepthTexture, texCoord).r * 2 - 1; // range between [-1, 1]
	vec2 xy = texCoord * 2 - 1;
	vec4 projectedPos = vec4(xy, z, 1.0f);

	vec4 pos_world_hom = InvViewProjMatrix * projectedPos;
	return pos_world_hom.xyz / pos_world_hom.w;
}

void main()
{
	ivec2 screenSize = textureSize(DiffuseTexture, 0); // size of GBuffer = size of screen
	vec2 texCoord = gl_FragCoord.xy / screenSize;

	vec3 position_world = positionFromDepth(texCoord);
	vec3 normal_world = texture(NormalTexture, texCoord).xyz;
	float specularIntensity = texture(DiffuseTexture, texCoord).w;
	float specularPower = clamp(texture(MaterialTexture, texCoord).r * 65536.0f, 1.0f, 100.0f);

	vec3 L_world = LightPosition_world - position_world;
	vec3 L = normalize(L_world);
	vec3 V = normalize( EyePosition_world - position_world );
	vec3 H = normalize( V + L );
	float NL = clamp(dot(normal_world, L), 0.0f, 1.0f);
	float NH = clamp(dot(normal_world, H), 0.0f, 1.0f);

	float lightOcclusion = 1;
	if (IncludeAmbientOcclusion)
	{
		vec4 occlusion = texture(OcclusionTexture, texCoord);
		lightOcclusion = 1 - clamp(dot(vec4(L, 1.0f), occlusion), 0.0f, 1.0f); // 1 if no occlusion
	}

	float distL = length(L_world);
	float attenuation = 1.0f / dot( LightAttenuation, vec3(1.0f, distL, distL * distL) );
	attenuation = min(1.0f, attenuation); // avoid wrong settings to increase light instead of attenuate
	attenuation *= step(distL, LightRange); // is zero if LightRange < distance to light

	// See also http://www.realtimerendering.com/blog/deferred-lighting-approaches/ for details
	
	
	vec3 BL = LightColor * LightPower * attenuation * lightOcclusion;

	// Diffuse color (RGB)
	vec3 diffuse = BL * NL;

	// Specular intensity (for alpha channel)
	vec3 specular = BL * specularIntensity * pow( NH, specularPower ); // TODO Use Fresnel term
	float specularLum = dot(vec3(0.2126f, 0.7152f, 0.0722f), specular);

	FragColor = vec4(diffuse, specularLum);
}
