#version 420 core

struct PatchData
{
	// control points:
	vec3 b210;
	vec3 b120;
	vec3 b021;
	vec3 b012;
	vec3 b102;
	vec3 b201;
	vec3 b111;
	
	// normals:
	vec3 n110;
	vec3 n011;
	vec3 n101;

	// tangents:
	//vec3 t110;
	//vec3 t011;
	//vec3 t101;
};


// define the number of CPs in the output patch
layout (vertices = 3) out;

uniform vec3  EyePosition_world;
uniform mat4  CameraMatrix;
uniform bool  OrientationAdaptive = true;
uniform bool  DistanceAdaptive = true;
uniform float MaximumTesselationLevel = 50.0f;


// attributes of the input CPs:
in vec3 Position_world0[];
in vec2 TexCoord0[];
in vec3 Normal_world0[];
in vec3 Tangent_world0[];

// output:
patch out PatchData PatchData1;
out vec2 TexCoord1[];
out vec3 Position_world1[];
out vec3 Normal_world1[];
out vec3 Tangent_world1[];

float CalcDistanceAdaptiveScale(vec3 point1, vec3 point2, float nearClip, float farClip)
{
	vec3 midPoint = (point1 + point2) * 0.5f; // middle point on edge

	// tessellation scaling factor [0,1] (if zero -> no tessellation, if one -> max tessellation)
	return 1.0f - clamp( (distance(midPoint, EyePosition_world) - nearClip) / (farClip - nearClip), 0.0f, 1.0f );
}

float CalcOrientationAdaptiveScale(vec3 normal1, vec3 normal2, vec3 viewVector, float silhouetteFactor)
{
	vec3 normal = normalize((normal1 + normal2) * 0.5f); // average normal

	// tessellation scaling factor [0,1] (if zero -> no tessellation, if one -> max tessellation)
	return pow(1.0f - abs( dot( normal, viewVector ) ), silhouetteFactor);
}

void QuadraticPatchPoints(in vec3 P0, in vec3 P1, in vec3 P2, in vec3 N0, in vec3 N1, in vec3 N2, out vec3 O1, out vec3 O2, out vec3 O3)
{
	// @see PN_Triangles - Section 3.3:

	vec3 P01 = P1 - P0;
	vec3 P12 = P2 - P1;
	vec3 P02 = P2 - P0;
	vec3 N0p1 = N0 + N1;
	vec3 N1p2 = N1 + N2;
	vec3 N0p2 = N0 + N2;

	float v12 = 2.0f * dot( P01, N0p1 ) / dot( P01, P01 );
	vec3 h110 = N0p1 - v12 * P01;
	O1 = normalize(h110);

	float v23 = 2.0f * dot( P12, N1p2 ) / dot( P12, P12 );
	vec3 h011 = N1p2 - v23 * P12;
	O2 = normalize(h011);

	float v31 = 2.0f * dot( P02, N0p2 ) / dot( P02, P02 );
	vec3 h101 = N0p2 - v31 * P02;
	O3 = normalize(h101);
}

void main()
{
	// gl_InvocationID contains the index of the current invocation within this patch

	// pass through for each invocation id:
	TexCoord1[gl_InvocationID] = TexCoord0[gl_InvocationID];
	Position_world1[gl_InvocationID] = Position_world0[gl_InvocationID];
	Normal_world1[gl_InvocationID] = Normal_world0[gl_InvocationID];
	Tangent_world1[gl_InvocationID] = Tangent_world0[gl_InvocationID];

	// here all invocations of one patch act the same:
	vec3 P0 = Position_world0[0];
	vec3 P1 = Position_world0[1];
	vec3 P2 = Position_world0[2];

	vec3 N0 = Normal_world0[0];
	vec3 N1 = Normal_world0[1];
	vec3 N2 = Normal_world0[2];

	vec3 T0 = Tangent_world0[0];
	vec3 T1 = Tangent_world0[1];
	vec3 T2 = Tangent_world0[2];

	// edge control points
	vec3 b210 = ( ( 2.0f * P0 ) + P1 - ( dot( ( P1 - P0 ), N0 ) * N0 ) ) / 3.0f;
	vec3 b120 = ( ( 2.0f * P1 ) + P0 - ( dot( ( P0 - P1 ), N1 ) * N1 ) ) / 3.0f;
	vec3 b021 = ( ( 2.0f * P1 ) + P2 - ( dot( ( P2 - P1 ), N1 ) * N1 ) ) / 3.0f;
	vec3 b012 = ( ( 2.0f * P2 ) + P1 - ( dot( ( P1 - P2 ), N2 ) * N2 ) ) / 3.0f;
	vec3 b102 = ( ( 2.0f * P2 ) + P0 - ( dot( ( P0 - P2 ), N2 ) * N2 ) ) / 3.0f;
	vec3 b201 = ( ( 2.0f * P0 ) + P2 - ( dot( ( P2 - P0 ), N0 ) * N0 ) ) / 3.0f;
	
	// center control point
	vec3 e = ( b210 + b120 + b021 + b012 + b102 + b201 ) / 6.0f;
	vec3 v = ( P0 + P1 + P2 ) / 3.0f;
	vec3 b111 = e + ( ( e - v ) / 2.0f );

	PatchData1.b210 = b210;
	PatchData1.b120 = b120;
	PatchData1.b021 = b021;
	PatchData1.b012 = b012;
	PatchData1.b102 = b102;
	PatchData1.b201 = b201;
	PatchData1.b111 = b111;

	// Calculate quadratic control points (for normals, tangents, bitangents)
	vec3 n110, n011, n101;
	QuadraticPatchPoints(P0, P1, P2, N0, N1, N2, n110, n011, n101);
	PatchData1.n110 = n110;
	PatchData1.n011 = n011;
	PatchData1.n101 = n101;

	//vec3 t110, t011, t101;
	//QuadraticPatchPoints(P0, P1, P2, T0, T1, T2, t110, t011, t101);
	//PatchData1.t110 = t110;
	//PatchData1.t011 = t011;
	//PatchData1.t101 = t101;

	// Calculate tessellation levels:
	gl_TessLevelOuter[0] = MaximumTesselationLevel;
	gl_TessLevelOuter[1] = MaximumTesselationLevel;
	gl_TessLevelOuter[2] = MaximumTesselationLevel;

	if (OrientationAdaptive)
	{
		// orientation adaptive tessellation
		// @see http://developer.download.nvidia.com/presentations/2010/gdc/Tessellation_Performance.pdf


		// View vector = vector from 'eye point' to 'look-at' point
		// i.e. the negative Z axis of the camera matrix:
		vec3 viewVector = normalize(CameraMatrix * vec4(0, 0, 1, 0)).xyz;

		// Silhouette factor. Weakens the dot product.
		// if normal and view vector are parallel -> dot product = 1 -> scale = 0 -> tessellation = lowest.
		const float silhouetteFactor = 0.5; // smaller factor -> reaches faster 

		float scale = CalcOrientationAdaptiveScale(N2, N0, viewVector, silhouetteFactor);
		gl_TessLevelOuter[0] = min(gl_TessLevelOuter[0], mix(3.0f, MaximumTesselationLevel, scale));

		scale = CalcOrientationAdaptiveScale(N0, N1, viewVector, silhouetteFactor);
		gl_TessLevelOuter[1] = min(gl_TessLevelOuter[1], mix(3.0f, MaximumTesselationLevel, scale));

		scale = CalcOrientationAdaptiveScale(N1, N2, viewVector, silhouetteFactor);
		gl_TessLevelOuter[2] = min(gl_TessLevelOuter[2], mix(3.0f, MaximumTesselationLevel, scale));
	}

	if (DistanceAdaptive)
	{
		// distance adaptive tessellation
		// @see https://developer.nvidia.com/content/dynamic-hardware-tessellation-basics

		const float nearClip = 30.0f; // max tessellation reached at this distance and if any closer
		const float farClip = 150.0f; // min tessellation reached at this distance and if further away

		float scale = CalcDistanceAdaptiveScale(P2, P0, nearClip, farClip);
		gl_TessLevelOuter[0] = min(gl_TessLevelOuter[0], mix(1.0f, MaximumTesselationLevel, scale));

		scale = CalcDistanceAdaptiveScale(P0, P1, nearClip, farClip);
		gl_TessLevelOuter[1] = min(gl_TessLevelOuter[1], mix(1.0f, MaximumTesselationLevel, scale));

		scale = CalcDistanceAdaptiveScale(P1, P2, nearClip, farClip);
		gl_TessLevelOuter[2] = min(gl_TessLevelOuter[2], mix(1.0f, MaximumTesselationLevel, scale));
	}

	// tessellation inside is the average of edge tessellations:
	// no right answer! could be anything (max, min, constant, ...)
	gl_TessLevelInner[0] = (gl_TessLevelOuter[0] + gl_TessLevelOuter[1] + gl_TessLevelOuter[2]) / 3.0f;
}
