#version 420 core

struct PatchData
{
	// control points:
	vec3 b210;
	vec3 b120;
	vec3 b021;
	vec3 b012;
	vec3 b102;
	vec3 b201;
	vec3 b111;
	
	// normals:
	vec3 n110;
	vec3 n011;
	vec3 n101;

	// tangents:
	//vec3 t110;
	//vec3 t011;
	//vec3 t101;
};


layout(triangles, equal_spacing, ccw) in;

uniform mat4 VpMatrix;
uniform sampler2D DisplacementTexture;
uniform sampler2D NormalTexture;
uniform float DisplacementFactor;
uniform float DisplacementNeutral = 0;

patch in PatchData PatchData1;
in vec2 TexCoord1[];
in vec3 Position_world1[];
in vec3 Normal_world1[];
in vec3 Tangent_world1[];

out vec3 WorldPos0;
out vec2 TexCoord0;
out vec3 Normal0;
out vec3 Tangent0;


vec2 InterpolateLinear2(vec2 v0, vec2 v1, vec2 v2, vec3 uvw)
{
	return uvw.z * v0 + uvw.x * v1 + uvw.y * v2;
}

vec3 InterpolateLinear3(vec3 v0, vec3 v1, vec3 v2, vec3 uvw)
{
	return uvw.z * v0 + uvw.x * v1 + uvw.y * v2;
}

vec3 InterpolateQuadratic(vec3 n200, vec3 n020, vec3 n002, vec3 n110, vec3 n011, vec3 n101, vec3 uvw, vec3 uvwSquared)
{
	return	n200 * uvwSquared.z +
			n020 * uvwSquared.x +
			n002 * uvwSquared.y +
			n110 * uvw.z * uvw.x +
			n011 * uvw.x * uvw.y +
			n101 * uvw.z * uvw.y;
}

vec3 InterpolateTriNormals( vec3 P1, vec3 P2, vec3 P3, vec3 N1, vec3 N2, vec3 N3, vec3 barycentricCoords)
{
	float u = barycentricCoords.x;
	float v = barycentricCoords.y;
	float w = barycentricCoords.z;

	// @see PN_Triangles - Section 3.3:
	float v12 = 2.0f * dot( P2 - P1, N1 + N2 ) / dot( P2 - P1, P2 - P1 );
	vec3 h110 = N1 + N2 - v12 * (P2 - P1);
	vec3 f3N110 = normalize(h110);

	float v23 = 2.0f * dot( P3 - P2, N2 + N3 ) / dot( P3 - P2, P3 - P2 );
	vec3 h011 = N2 + N3 - v23 * (P3 - P2);
	vec3 f3N011 = normalize(h011);

	float v31 = 2.0f * dot( P3 - P1, N1 + N3 ) / dot( P3 - P1, P3 - P1 );
	vec3 h101 = N3 + N1 - v31 * (P3 - P1);
	vec3 f3N101 = normalize(h101);

	// @see PN_Triangles - Section 2:
	return		N1 * w * w +
				N2 * u * u +
				N3 * v * v +
				f3N110 * w * u +
				f3N011 * u * v +
				f3N101 * w * v;
}


void main()
{
	vec3 b300 = Position_world1[0]; // P1
	vec3 b030 = Position_world1[1]; // P2
	vec3 b003 = Position_world1[2]; // P3
	vec3 P1 = b300;	vec3 P2 = b030;	vec3 P3 = b003;

	vec3 uvw = gl_TessCoord;
	float u = uvw.x;
	float v = uvw.y;
	float w = uvw.z;

	vec3 uvwSquared = uvw * uvw;
	vec3 uvwCubed = uvwSquared * uvw;

	vec3 uvwSquared3 = uvwSquared * 3;

	// Cubic interpolation of position.
	// @see PN_Triangles - Section 2:

	vec3 position =			   b300 * uvwCubed.z +
							   b030 * uvwCubed.x +
							   b003 * uvwCubed.y +
					PatchData1.b210 * uvwSquared3.z * u +
					PatchData1.b120 * uvwSquared3.x * w  +
					PatchData1.b201 * uvwSquared3.z * v +
					PatchData1.b021 * uvwSquared3.x * v +
					PatchData1.b102 * uvwSquared3.y * w +
					PatchData1.b012 * uvwSquared3.y * u +
					PatchData1.b111 * 6 * w * u * v;

	// Quadratic interpolation of normals
	// @see PN_Triangles - Section 2:
	vec3 normal = InterpolateQuadratic(Normal_world1[0], Normal_world1[1], Normal_world1[2], PatchData1.n110, PatchData1.n011, PatchData1.n101, uvw, uvwSquared);

	// tangents will (in general) have discontinuities with quadratic interpolation
	// -> use simple linear interpolation:
	vec3 tangent = InterpolateLinear3(Tangent_world1[0], Tangent_world1[1], Tangent_world1[2], uvw);
	//vec3 tangent = InterpolateQuadratic(Tangent_world1[0], Tangent_world1[1], Tangent_world1[2], PatchData1.t110, PatchData1.t011, PatchData1.t101, uvw, uvwSquared);

	TexCoord0 = InterpolateLinear2(TexCoord1[0], TexCoord1[1], TexCoord1[2], uvw);
	Normal0 = normal;
	Tangent0 = tangent;

	// displacement mapping:
	float displacement = texture(DisplacementTexture, TexCoord0.xy).x - DisplacementNeutral;
	vec3 alongNormal = normalize(Normal0);

	WorldPos0 = position + alongNormal * displacement * DisplacementFactor;
	gl_Position = VpMatrix * vec4(WorldPos0, 1.0);
}
