#version 330

#include "colorSpaces.glsl"
#line 5
in vec2 TexCoord0;

out vec4 FragColor;

uniform sampler2D Texture0;
uniform float MaximumLuminance;
uniform float AverageLuminance;

float log10(float x)
{
	return log(x) / log(10.0f);
}

vec3 toneMapReinhard(vec3 color) 
{
	vec3 Yxy = rgb2Yxy(color);

	float L_p = Yxy.r;
	float L_d = L_p/(1.0f + L_p);
	
	Yxy.r = L_d;
	color.rgb = Yxy2rgb(Yxy);
	return color;
}

vec3 toneMapModifiedReinhard(vec3 color, float maximumLuminance) 
{
	float L_white = maximumLuminance;
	vec3 Yxy = rgb2Yxy(color);

	float L_p = Yxy.r; // luminance pixel
	float L_d = L_p * (1.0f + L_p / (L_white * L_white)) / (1.0f + L_p);
	
	Yxy.r = L_d;
	color.rgb = Yxy2rgb(Yxy);
	return color;
}

vec3 toneMapDrago(vec3 color, float maximumLuminance)
{
/*

L_w ... world luminance (pro pixel)
L_wmax ... maximum luminance in the scene
L_wa ... world adaptation luminance		(to ignore?)
L_d ... computed displaying value (per pixel)
L_dmax ... maximum luminance capability of the displaying medium (scale factor) (used: 100 cd/m^2, Notebook: 283 cd/m^2)
b ... bias. Proposed values between 0.5 and 1 (mit 0.73 ~ das gleiche Mapping wie mit gamma = 2.2)
		adjusts compression of high values and visibility of details in dark areas.
		According to user survey b = 0.85 was proposed

*/

	float L_wmax = maximumLuminance;
	const float b = 0.85f;
	const float L_dmax = 100.0f; // would actually depend on the displaying medium

	vec3 Yxy = rgb2Yxy(color);

	float L_w = Yxy.r;

	float term1 = L_dmax * 0.01 / log10(L_wmax + 1);
	float denominator = log( 2 + pow(L_w / L_wmax, log(b) / log(0.5)) * 8);
	float term2 = log(L_w + 1) / denominator;

	float L_d = term1 * term2;
	Yxy.r = L_d;

	color.rgb = Yxy2rgb(Yxy);
	return color;
}

vec3 curveUncharted2(vec3 x)
{
	// Information: http://www.slideshare.net/ozlael/hable-john-uncharted2-hdr-lighting (slide 142ff)
	// Also: http://www.slideshare.net/hpduiker/filmic-tonemapping-for-realtime-rendering-siggraph-2010-color-course (slide 56)

	const float ShoulderStrength = 0.22f; // 0.15f;
	const float LinearStrength = 0.30f; // 0.50f;
	const float LinearAngle = 0.10f;
	const float ToeStrength = 0.20f;
	const float ToeNumerator = 0.01f; // 0.02f;
	const float ToeDenominator = 0.30f;

	float A = ShoulderStrength;
	float B = LinearStrength;
	float C = LinearAngle;
	float D = ToeStrength;
	float E = ToeNumerator;
	float F = ToeDenominator;
	// E / F = Toe Angle

	return ((x*(A*x+C*B)+D*E)/(x*(A*x+B)+D*F)) - E/F;
}

// Applies the Uncharted 2 filmic tone mapping curve
vec3 toneMapUncharted2(vec3 color, float maximumLuminance)
{
	// LinearWhite ... linear white point value
	//const vec3 LinearWhite = vec3(11.2f);
	vec3 LinearWhite = vec3(maximumLuminance);

	vec3 numerator = curveUncharted2(color);
	vec3 denominator = curveUncharted2(LinearWhite);

	return numerator / denominator;
}

vec3 applyExposure(vec3 color, float avgLuminance)
{
	// automatic middle gray calculation:
	// http://content.gpwiki.org/index.php/D3DBook:High-Dynamic_Range_Rendering
	// (Equation 10)
	float tonemappingKey = 1.03f - (2.0f / (2.0f + log10(avgLuminance + 1.0f)));
	float exposure = (tonemappingKey / avgLuminance);

	return color * exposure;
}

void main()
{
	// for our scene we need to cap the luminance values, such that very dark parts are not totally overexposed
	float avgLuminance = max(0.03f, AverageLuminance);
	float maxLuminance = min(0.7f, MaximumLuminance); // capping the max luminance for overall brighter appearance (especially for beamer)
	//float maxLuminance = MaximumLuminance;

	vec3 color = texture(Texture0, TexCoord0).xyz;

	color = applyExposure(color, avgLuminance);

	//vec3 newColor = toneMapModifiedReinhard(color, maxLuminance);
	//vec3 newColor = toneMapDrago(color, maxLuminance);
	vec3 newColor = toneMapUncharted2(color, maxLuminance);
	newColor = max(newColor, vec3(0));
	FragColor = vec4(newColor, 1.0f);
}
